//! This module is for math utils missing from core.
//!
//! Over time, it's hoped that most of these will be added to core. Once that
//! happens they will be removed from here as part of some future braking change.

use super::*;

#[cfg(all(target_arch = "x86", target_feature = "sse"))]
use arch::x86::m128;
#[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
use arch::x86_64::m128;

#[cfg(all(target_arch = "x86", target_feature = "sse2"))]
use arch::x86::{m128d, m128i};
#[cfg(all(target_arch = "x86_64", target_feature = "sse2"))]
use arch::x86_64::{m128d, m128i};

/// `f32` absolute value.
#[inline(always)]
pub fn abs_f32(f: f32) -> f32 {
  // Note(Lokathor): Only tested on little-endian, but ScottMCM of T-lang claims
  // that this will be fine on big-endian machines so #TRUST
  f32::from_bits(f.to_bits() & (i32::max_value() as u32))
}

/// `f64` absolute value.
#[inline]
pub fn abs_f64(f: f64) -> f64 {
  // Note(Lokathor): Only tested on little-endian, but ScottMCM of T-lang claims
  // that this will be fine on big-endian machines so #TRUST
  f64::from_bits(f.to_bits() & (i64::max_value() as u64))
}

/// `f32` ceiling (round toward positive infinity).
#[inline(always)]
pub fn ceil_f32(f: f32) -> f32 {
  #[cfg(target_feature = "sse4.1")]
  {
    m128::set0(f).ceil().extract0()
  }
  #[cfg(not(target_feature = "sse4.1"))]
  {
    // Note(Lokathor): This magical value is the smallest `f32` that will never
    // have a fractional part (1*2^MANTISSA_BITS*1). Any number of this value or
    // more is already a whole number and we don't need to do anything. We
    // compare based on bit patterns so that +/- Inf and NaNs also fall into the
    // "do nothing" category.
    if abs_f32(f).to_bits() < 8_388_608.0_f32.to_bits() {
      let truncated = f as i32 as f32;
      if truncated < f {
        truncated + 1.0
      } else {
        truncated
      }
    } else {
      f
    }
  }
}

/// `f64` ceiling (round toward positive infinity).
#[inline(always)]
pub fn ceil_f64(f: f64) -> f64 {
  #[cfg(target_feature = "sse4.1")]
  {
    m128d::set0(f).ceil().extract0()
  }
  #[cfg(not(target_feature = "sse4.1"))]
  {
    // Note(Lokathor): See `ceil_f32`, this is just the f64 version of the same
    // concept.
    if abs_f64(f).to_bits() < 4_503_599_627_370_496.0_f64.to_bits() {
      let truncated = f as i64 as f64;
      if truncated < f {
        truncated + 1.0
      } else {
        truncated
      }
    } else {
      f
    }
  }
}

/// `f32` floor (round toward negative infinity).
#[inline(always)]
pub fn floor_f32(f: f32) -> f32 {
  #[cfg(target_feature = "sse4.1")]
  {
    m128::set0(f).floor().extract0()
  }
  #[cfg(not(target_feature = "sse4.1"))]
  {
    // Note(Lokathor): See `ceil_f32`, all that changes here is that we subtract
    // 1 instead of adding 1 if we need to do a change.
    if abs_f32(f).to_bits() < 8_388_608.0_f32.to_bits() {
      let truncated = f as i32 as f32;
      if truncated > f {
        truncated - 1.0
      } else {
        truncated
      }
    } else {
      f
    }
  }
}

/// `f64` floor (round toward negative infinity).
#[inline(always)]
pub fn floor_f64(f: f64) -> f64 {
  #[cfg(target_feature = "sse4.1")]
  {
    m128d::set0(f).floor().extract0()
  }
  #[cfg(not(target_feature = "sse4.1"))]
  {
    // Note(Lokathor): See `ceil_f32`, this is just the f64 version of the same
    // concept. All that changes here is that we subtract 1 instead of adding 1
    // if we need to do a change.
    if abs_f64(f).to_bits() < 4_503_599_627_370_496.0_f64.to_bits() {
      let truncated = f as i64 as f64;
      if truncated > f {
        truncated - 1.0
      } else {
        truncated
      }
    } else {
      f
    }
  }
}

/// `f32` square root.
///
/// short-circuits to NaN if the input is < 0.0
#[inline(always)]
pub fn sqrt_f32(f: f32) -> f32 {
  if f < 0.0 {
    return core::f32::NAN;
  }
  #[cfg(target_feature = "sse")]
  {
    m128::set0(f).sqrt0().extract0()
  }
  #[cfg(not(target_feature = "sse"))]
  {
    libm::sqrtf(f)
  }
}

/// `f64` square root.
///
/// short-circuits to NaN if the input is < 0.0
#[inline]
pub fn sqrt_f64(f: f64) -> f64 {
  if f < 0.0 {
    return core::f64::NAN;
  }
  #[cfg(target_feature = "sse2")]
  {
    m128d::set0(f).sqrt().extract0()
  }
  #[cfg(not(target_feature = "sse2"))]
  {
    libm::sqrt(f)
  }
}
